open Printf
open MIPSOps
open Primitive
open InterpretUPP
open InterpretRTL
open ERTL
open Integer

(* TEMPORARY this interpreter will fail mysteriously if the code sets [$ra] incorrectly;
             it will complain that pseudo-registers are being read from the wrong universe.
             What can we do about that? *)

(* ------------------------------------------------------------------------- *)

(* This exception is intentionally never caught. This allows us to obtain
   stack backtraces when debugging. *)

exception RuntimeError = InterpretPP.RuntimeError

(* ------------------------------------------------------------------------- *)

(* In [ERTL], an address is either a pair of a procedure and a label
   or the initial address (used to start the interpreter). *)

type ('procedure, 'label) address =
  | AddrCode of 'procedure * 'label
  | AddrInit

(* ------------------------------------------------------------------------- *)

(* Accessing hardware registers. *)

type 'value henv =
    'value ref MIPS.RegisterMap.t

let hlookup henv hwr =
  try
    MIPS.RegisterMap.find hwr henv
  with Not_found ->
    assert false

let hread henv hwr =
  if MIPS.equal hwr MIPS.zero then
    VInt 0l (* reading register [$zero] always reads 0 *)
  else
    !(hlookup henv hwr)

let hwrite henv hwr v =
  (hlookup henv hwr) := v

(* ------------------------------------------------------------------------- *)

(* Interpreting primitive operations. Parameters and results are exchanged
   in hardware registers. *)

let interpret_primitive henv = function
  | Write ->
      let v = hread henv (List.hd MIPS.parameters) in
      fprintf stdout "%ld%!" (asInt v)
  | Writeln ->
      let v = hread henv (List.hd MIPS.parameters) in
      fprintf stdout "%ld\n%!" (asInt v)
  | Readln ->
      let v = VInt (Int32.of_string (input_line stdin)) in
      hwrite henv MIPS.result v
  | Alloc ->
      let v = hread henv (List.hd MIPS.parameters) in
      hwrite henv MIPS.result (interpret_alloc v)

(* ------------------------------------------------------------------------- *)

(* Accessing stack slots. The incoming and outgoing stack areas are
   represented as mappings of offsets to value references. Slots in
   the outgoing area are allocated on demand, since the stack can grow
   indefinitely. The size of the incoming area is fixed. The local
   area is represented as a mapping of pseudo-registers to value
   references.

   The semantics of [ERTL] specifies that [INewFrame] pushes a stack
   frame while [IDeleteFrame] pops one, thus restoring the caller's
   frame.

   Note that [IReturn] jumps to the address specified by [$ra], which
   is not necessarily the caller's address if [$ra] was not correctly
   set. *)

module Int32Map =
  Map.Make(Int32)

type frame = {
    incoming: ertl_address value ref Int32Map.t;
    mutable outgoing: ertl_address value ref Int32Map.t;
    local: ertl_address value ref Register.Map.t
  }

and ertl_address =
    (ERTL.procedure, Label.t) address

type stack =
    frame list

let slookup stack slot =
  let frame = List.hd stack in
  match slot with
  | SlotIncoming offset ->
      begin try
	Int32Map.find offset frame.incoming
      with Not_found ->
	fprintf stderr "Runtime error -- invalid incoming slot (%ld).\n" offset;
	raise RuntimeError
      end
  | SlotOutgoing offset ->
      begin try
	Int32Map.find offset frame.outgoing
      with Not_found ->
	let cell = ref default in
	frame.outgoing <- Int32Map.add offset cell frame.outgoing;
	cell
      end

let read stack r =
  let frame = List.hd stack in
  read frame.local r

let write stack r v =
  let frame = List.hd stack in
  write frame.local r v

let sread stack slot =
  !(slookup stack slot)

let swrite stack slot v =
  (slookup stack slot) := v

(* An empty stack. *)

let empty_stack =
  []

(* Creating a stack frame. The callee's incoming stack area is the
   caller's outgoing stack area. The callee's outgoing stack area is
   empty. The callee's frame is pushed on top of the stack. *)

let newframe stack locals =
  match stack with
  | [] ->
      [{
         incoming = Int32Map.empty;
         outgoing = Int32Map.empty;
         local = Register.Map.lift allocate locals
      }]
  | frame :: _ ->
      {
        incoming = frame.outgoing;
        outgoing = Int32Map.empty;
         local = Register.Map.lift allocate locals
      } :: stack

(* Deleting a stack frame. We assume that incoming parameters are
   never written and outgoing parameters are never read, so that the
   caller's stack is unaffected by the call. *)

let deleteframe stack =
  List.tl stack

(* ------------------------------------------------------------------------- *)

(* Interpreting programs. *)

let interpret p =

  (* Create an environment [genv] that holds the global variables. *)

  let genv = Array.make (p.globals / MIPS.word) (VInt 0l) in

  (* Create an environment [henv] that holds the hardware registers. *)

  let henv = MIPS.RegisterMap.lift allocate MIPS.allocatable in

  (* The code that follows refers to [p.defs], [genv], and [henv]
     where necessary. *)

  (* ----------------------------------------------------------------------- *)

  (* Interpreting function and procedure calls. *)

  let rec interpret_call stack callee next =
    match callee with

    | CPrimitiveFunction p ->

	interpret_primitive henv p;
	interpret_address stack next

    | CUserFunction f ->

	(* Lookup the definition of the procedure [f]. *)

	let proc =
	  try
	    StringMap.find f p.defs
	  with Not_found ->
	    assert false
	in

	(* Write the address of the next instruction into [$ra], push a
	   new stack frame, and execute the procedure body. *)

	hwrite henv MIPS.ra (VCode next);
	interpret_code stack proc proc.entry

  (* ----------------------------------------------------------------------- *)

  (* Interpreting addresses. *)

  and interpret_address stack = function
    | AddrInit ->

	(* Halt. *)

	()

    | AddrCode (proc, l) ->

	interpret_code stack proc l

  and interpret_code stack proc l =

    (* Fetch instruction. *)

    let i =
      try
	Label.Map.find l proc.graph
      with Not_found ->
	fprintf stderr
	  "Runtime error -- no instruction is associated with the current label (%s).\n" (Label.print l);
	raise RuntimeError
    in

    (* Execute. *)

    interpret_instruction stack proc i

  (* ----------------------------------------------------------------------- *)

  (* Interpreting instructions. *)

  and interpret_instruction stack proc i =
    match i with

    | INewFrame l ->
	interpret_code (newframe stack proc.locals) proc l

    | IDeleteFrame l ->
	interpret_code (deleteframe stack) proc l

    | IGetHwReg (destr, sourcehwr, l) ->
	write stack destr (hread henv sourcehwr);
	interpret_code stack proc l

    | ISetHwReg (desthwr, sourcer, l) ->
	hwrite henv desthwr (read stack sourcer);
	interpret_code stack proc l

    | IGetStack (destr, slot, l) ->
	write stack destr (sread stack slot);
	interpret_code stack proc l

    | ISetStack (slot, sourcer, l) ->
	swrite stack slot (read stack sourcer);
	interpret_code stack proc l

    | IConst (destr, i, l) ->
	write stack destr (VInt i);
	interpret_code stack proc l

    | IUnOp (op, destr, sourcer, l) ->
	write stack destr (unop op (read stack sourcer));
	interpret_code stack proc l

    | IBinOp (op, destr, sourcer1, sourcer2, l) ->
	write stack destr (binop op (read stack sourcer1) (read stack sourcer2));
	interpret_code stack proc l

      (* Pass the address of the next instruction as the return address. *)

    | ICall (callee, _, l) ->
	interpret_call stack callee (AddrCode (proc, l))

    | ILoad (destr, addressr, offset, l) ->
	write stack destr (load (read stack addressr) offset);
	interpret_code stack proc l

    | IStore (addressr, offset, valuer, l) ->
	store (read stack addressr) offset (read stack valuer);
	interpret_code stack proc l

    | IGetGlobal (destr, offset, l) ->
	write stack destr (gread genv offset);
	interpret_code stack proc l

    | ISetGlobal (offset, sourcer, l) ->
	gwrite genv offset (read stack sourcer);
	interpret_code stack proc l

    | IGoto l ->
	interpret_code stack proc l

    | IUnBranch (cond, sourcer, l1, l2) ->
	interpret_code stack proc
	  (if InterpretMIPS.uncon cond (asInt (read stack sourcer)) then l1 else l2)

    | IBinBranch (cond, sourcer1, sourcer2, l1, l2) ->
	interpret_code stack proc
	  (if InterpretMIPS.bincon cond (asInt (read stack sourcer1)) (asInt (read stack sourcer2)) then l1 else l2)

      (* Jump to [$ra]. *)

    | IReturn _ ->
	interpret_address stack (asAddress (hread henv MIPS.ra))

  in

  (* ----------------------------------------------------------------------- *)

  (* Interpreting programs (end). *)

  (* Execute the program body. *)

  interpret_call empty_stack (CUserFunction "_main") AddrInit

