module Make (Env : sig

  (* [lookup x] returns the pseudo-register that holds the local
     variable [x]. *)

  val lookup: string -> Register.t

  (* [allocate()] returns a fresh pseudo-register. *)

  val allocate: unit -> Register.t

  (* [generate instruction] returns a fresh instruction label, which
     it associates with [instruction] in the control flow graph. *)

  val generate: RTL.instruction -> Label.t

  (* [loop target] returns a fresh instruction label [label], which it
     associates with an unconditional branch instruction whose target
     is [target label]. *)

  val loop: (Label.t -> Label.t) -> Label.t

end) = struct

open Integer
open MIPSOps
open Env

(* ------------------------------------------------------------------------- *)

(* Translating expressions. *)

(* [translate_expression destr e destl] generates new instructions
   whose effect is to evaluate the expression [e], to place its value
   in the pseudo-register [destr], and to transfer control to the
   label [destl]. It returns the entry label of the newly generated
   instructions. *)

let rec translate_expression
   (destr : Register.t)
   (e : UPP.expression)
   (destl : Label.t)
    : Label.t =

  match e with

    (* Constants. *)

  | UPP.EConst i -> failwith "Pas implment !"

    (* Local variable access. *)

  | UPP.EGetVar x -> failwith "Pas implment !"

    (* Global variable access. *)

  | UPP.EGetGlobal offset -> failwith "Pas implment !"

    (* Unary operator applications. *)

  | UPP.EUnOp (op, e) -> failwith "Pas implment !"

    (* Binary operator applications. *)

  | UPP.EBinOp (op, e1, e2) -> failwith "Pas implment !"

    (* Function calls. *)

  | UPP.EFunCall (callee, actuals) -> failwith "Pas implment !"

    (* Memory reads. *)

  | UPP.ELoad (e, offset) -> failwith "Pas implment !"

(* Translating function and procedure calls. 
   The destination register is optional --
   present for functions, absent for procedures. *)

and translate_call 
   (destr : Register.t option)
   (callee : Primitive.callee)
   (actuals : UPP.expression list)
   (destl : Label.t)
   : Label.t =

  let temporaries = List.map (fun _ -> allocate()) actuals in
  List.fold_right2 translate_expression temporaries actuals (
  generate (RTL.ICall (destr, callee, temporaries, destl))
  )

(* ------------------------------------------------------------------------- *)

(* [mkunbranch e uncon truel falsel] translates the expression [e],
   writing its value to a temporary register; then, it issues a unary
   branch instruction on that register, whose condition is [uncon],
   and whose target labels are [truel] and [falsel]. *)

let mkunbranch e uncon truel falsel =
  let temporary = allocate() in
  translate_expression temporary e (
  generate (RTL.IUnBranch (uncon, temporary, truel, falsel))
  )

(* [mkbinbranch e1 e2 bincon truel falsel] translates the expressions
   [e1] and [e2], writing their values to temporary registers; then,
   it issues a binary branch instruction on those registers, whose
   condition is [bincon], and whose target labels are [truel] and
   [falsel]. *)

let mkbinbranch e1 e2 bincon truel falsel =
  let temporary1 = allocate()
  and temporary2 = allocate() in
  translate_expression temporary1 e1 (
  translate_expression temporary2 e2 (
  generate (RTL.IBinBranch (bincon, temporary1, temporary2, truel, falsel))
  ))

(* Translating conditions. *)

(* [translate_condition c truel falsel] generates new [RTL]
   instructions whose effect is to evaluate the condition [c] and to
   transfer control to one of the labels [truel] and [falsel],
   depending on the condition's value. It returns the entry label of
   the newly generated instructions. *)

let rec translate_condition
   (c : UPP.condition)
   (truel : Label.t)
   (falsel : Label.t)
    : Label.t =

  match c with

  | UPP.CExpression e -> failwith "Pas implment !"

    (* Boolean negation. *)

  | UPP.CNot c -> failwith "Pas implment !"

    (* Boolean conjunction. *)

  | UPP.CAnd (c1, c2) -> failwith "Pas implment !"

    (* Boolean disjunction. *)

  | UPP.COr (c1, c2) -> failwith "Pas implment !"

(* ------------------------------------------------------------------------- *)

(* Translating instructions. *)

(* [translate_instruction i destl] generates new [RTL] instructions
   whose effect is to execute the [UPP] instruction [i] and to
   transfer control to the destination label [destl]. It returns the
   entry label of the newly generated instructions. *)

let rec translate_instruction
   (i : UPP.instruction)
   (destl : Label.t)
    : Label.t =

  match i with

    (* Procedure calls. *)

  | UPP.IProcCall (callee, actuals) ->
      translate_call None callee actuals destl

    (* Local variable update. *)

  | UPP.ISetVar (x, e) -> failwith "Pas implment !"

    (* Global variable update. *)

  | UPP.ISetGlobal (offset, e) ->
      let temporary = allocate() in
      translate_expression temporary e (
        generate (RTL.ISetGlobal (offset, temporary, destl))
      )

    (* Memory write. This is translated to the corresponding [RTL]
       instruction. *)

  | UPP.IStore (eaddress, offset, evalue) ->
      let address = allocate()
      and value = allocate() in
      translate_expression address eaddress (
      translate_expression value evalue (
      generate (RTL.IStore (address, offset, value, destl))
      ))

    (* Sequence. This is translated by chaining the destination
       labels. *)

  | UPP.ISeq instructions ->
      List.fold_right translate_instruction instructions destl

    (* Conditional. *)

  | UPP.IIf (c, i1, i2) -> failwith "Pas implment !"

    (* Loop. *)

  | UPP.IWhile (c, i) -> failwith "Pas implment !"

(* ------------------------------------------------------------------------- *)

end

